<?php

class TruevoAPI
{
    /**
     * The root module.
     *
     * @var TruevoPayments
     */
    private $module;

    /**
     * A unique value generated to recognize subsequent retries of the same request.
     *
     * @var string
     */
    private $idempotencyKey;

    /**
     * The order saved in the store to match up with Truevo transaction
     * 
     * @var Order
     */
    private $order;

    /**
     * The constructor for the API handler.
     * 
     * @param Order $order
     */
    public function __construct(Order $order)
    {
        $this->module = Module::getInstanceByName('truevopayments');
        $this->idempotencyKey = $this->generateUuid();
        $this->order = $order;
    }

    /**
     * Generate a valid RFC 4211 Compliant version 4 Universally Unique Identifier.
     * 
     * @return string
     */
    private function generateUuid() 
    {
	    return sprintf( '%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
	        // 32 bits for "time_low"
	        mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff ),
	
	        // 16 bits for "time_mid"
	        mt_rand( 0, 0xffff ),
	
	        // 16 bits for "time_hi_and_version",
	        // four most significant bits holds version number 4
	        mt_rand( 0, 0x0fff ) | 0x4000,
	
	        // 16 bits, 8 bits for "clk_seq_hi_res",
	        // 8 bits for "clk_seq_low",
	        // two most significant bits holds zero and one for variant DCE1.1
	        mt_rand( 0, 0x3fff ) | 0x8000,
	
	        // 48 bits for "node"
	        mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff )
	    );
	}

    /**
     * Get the details of a transaction for refunds and reversals.
     *
     * @return string|false
     */
    public function getDetails()
    {
        $transactionDetails = $this->module->getTransactionDetailsForOrder($this->order);

        if (!$transactionDetails) {
            $this->module->truevoLogger('Failed to find Truevo transaction details for capturing payment on Order #' . $this->order->id);
            return false;
        }

        $data = [
			'transactionUniqueIdentifier' => $transactionDetails['transaction_id']
		];

        $response = $this->sendRequest('GET', 'payments/details', $data);

        // Build missing properties
        if ($response) {
            if ($response->outcome->status === "COMPLETE") {
                if (!property_exists($response->paymentDetails, 'isCaptureAllowed')) $response->paymentDetails->isCaptureAllowed = false;
                if (!property_exists($response->paymentDetails, 'isRefundAllowed')) $response->paymentDetails->isRefundAllowed = false;
                if (!property_exists($response->paymentDetails, 'isReverseAllowed')) $response->paymentDetails->isReverseAllowed = false;
                if (!property_exists($response->paymentDetails, 'availableForCapture')) $response->paymentDetails->availableForCapture = 0;
                if (!property_exists($response->paymentDetails, 'availableForRefund')) $response->paymentDetails->availableForRefund = 0;

                return $response;
            } else {
                $message = 'Unable to find the following Transaction ID on the system: ' . $data['transactionUniqueIdentifier'];
                $this->module->assignStatusNotification($this->order->id, 'warning', $message);
            }

            return false;
        }

        $this->module->assignStatusNotification($this->order->id, 'danger', 'No response querying the status of the following Transaction ID on the system: ' . $data['transactionUniqueIdentifier']);
        return false;
    }

    /**
     * Get the status of a transaction (to be used after a capture payment call).
     *
     * @param string $transactionUniqueIdentifier The unique identifier to check (A capture request counts as a different transaction and will return a new identifier).
     * @return string|false
     */
    public function getStatus($transactionUniqueIdentifier)
    {
        $data = [
			'transactionUniqueIdentifier' => $transactionUniqueIdentifier,
            'merchantIdentifier'          => $this->module->getMerchantMid(),
			'terminalIdentifier'          => $this->module->getMerchantTid(),
		];

        return $this->sendRequest('GET', 'payments/status', $data);
    }

    /**
     * Capture a payment for an order currently "On Hold".
     *
     * @return string|false
     */
    public function capturePayment()
    {
        $transactionDetails = $this->module->getTransactionDetailsForOrder($this->order);

        if (!$transactionDetails) {
            $this->module->truevoLogger('Failed to find Truevo transaction details for capturing payment on Order #' . $this->order->id);
            return false;
        }
        
        $currency = new Currency($this->order->id_currency);
        $customer = new Customer($this->order->id_customer);
        $address  = new Address($this->order->id_address_invoice);
        $state    = new State($address->id_state);
        $country  = new Country($address->id_country);

        $data = [
            "transaction" => [
                'originalTransactionUniqueIdentifier' => $transactionDetails['transaction_id'],
				'retrievalReferenceNumber' => $transactionDetails['rrn'],
                "paymentReference" => $transactionDetails['payment_reference'],
                "merchantReference" => $transactionDetails['merchant_reference'],
                "amount" => [
                    "totalAmount" => (float) $this->order->total_paid,
                    "currencyAlphaCode" => $currency->iso_code
                ],
                "source" => 'ECOMMERCE'
            ],
            "merchant" => [
                "mid" => $this->module->getMerchantMid(),
                "tid" => $this->module->getMerchantTid(),
            ],
            "customer" => [
                "firstName" => $customer->firstname,
                "lastName" => $customer->lastname,
                "birthDate" => $customer->birthday === "0000-00-00" ? null : $customer->birthday,
                "contact" => [
                    "email" => $customer->email,
                    "mobile" => $address->phone_mobile,
                    "phone" => $address->phone,
                    "ip" => $this->module->getClientIpAddress(),
                ],
                "billingAddress" => [
                    "street1" => $address->address1,
                    "street2" => $address->address2,
                    "city" => $address->city,
                    "state" => $state->name,
                    "postCodeZip" => $address->postcode,
                    "countryAlphaCode" => $country->iso_code,
                ]
            ],
            "additionalData" => []
        ];

        return $this->sendRequest('POST', 'payments/capture', $data);
    }

    /**
     * Send through a partial refund or a full refund request for an order.
     *
     * @param float|null $amount The amount to refund.
     * @param integer|null $numberOfPieces The number of products being returned.
     * @param string|null $reason The reason for the refund.
     * @return string|false
     */
    public function refund($amount = null, $numberOfPieces = null, $reason = null)
    {
        $transactionDetails = $this->module->getTransactionDetailsForOrder($this->order);

        if (!$transactionDetails) {
            $this->module->truevoLogger('Failed to find Truevo transaction details for refund on Order #' . $this->order->id);
            return false;
        }
        
        $additionalData = array();

        if (!is_null($numberOfPieces)) {
            $additionalData[] = [
                'key' => 'number_of_pieces',
                'value' => $numberOfPieces
            ];
        }

        if (!is_null($reason) && !empty($reason)) {
            $additionalData[] = [
                'key' => 'reason_for_refund',
                'value' => $reason
            ];
        }
        
        $totalAmount = !is_null($amount) ? $amount : $this->order->total_paid;

        $currency = new Currency($this->order->id_currency);
        $customer = new Customer($this->order->id_customer);
        $address  = new Address($this->order->id_address_invoice);
        $state    = new State($address->id_state);
        $country  = new Country($address->id_country);

        $data = [
            "transaction" => [
                'originalTransactionUniqueIdentifier' => $transactionDetails['transaction_id'],
                "paymentReference" => $transactionDetails['payment_reference'],
                "merchantReference" => $transactionDetails['merchant_reference'],
                "amount" => [
                    "totalAmount" => (float) $totalAmount,
                    "currencyAlphaCode" => $currency->iso_code
                ],
                "source" => 'ECOMMERCE'
            ],
            "merchant" => [
                "mid" => $this->module->getMerchantMid(),
                "tid" => $this->module->getMerchantTid(),
            ],
            "customer" => [
                "firstName" => $customer->firstname,
                "lastName" => $customer->lastname,
                "birthDate" => $customer->birthday === "0000-00-00" ? null : $customer->birthday,
                "contact" => [
                    "email" => $customer->email,
                    "mobile" => $address->phone_mobile,
                    "phone" => $address->phone,
                    "ip" => $this->module->getClientIpAddress(),
                ],
                "billingAddress" => [
                    "street1" => $address->address1,
                    "street2" => $address->address2,
                    "city" => $address->city,
                    "state" => $state->name,
                    "postCodeZip" => $address->postcode,
                    "countryAlphaCode" => $country->iso_code,
                ]
            ],
            "additionalData" => $additionalData
        ];

        return $this->sendRequest('POST', 'payments/linkedRefund', $data);
    }

    /**
     * Void the transaction and reverse it's effects. 
     * This operation can only be performed until a cut-off point for the transaction.
     *
     * @return string|false
     */
    public function reversal()
    {
        $transactionDetails = $this->module->getTransactionDetailsForOrder($this->order);

        if (!$transactionDetails) {
            $this->module->truevoLogger('Failed to find Truevo transaction details for reversal on Order #' . $this->order->id);
            return false;
        }
        
        $currency = new Currency($this->order->id_currency);
        $customer = new Customer($this->order->id_customer);
        $address  = new Address($this->order->id_address_invoice);
        $state    = new State($address->id_state);
        $country  = new Country($address->id_country);

        $data = [
            "transaction" => [
                'originalTransactionUniqueIdentifier' => $transactionDetails['transaction_id'],
                "paymentReference" => $transactionDetails['payment_reference'],
                "merchantReference" => $transactionDetails['merchant_reference'],
                'transactionType' => $transactionDetails['transaction_type'] === "AUTHORIZE" ? "CAPTURE" : "SALE",
                "source" => 'ECOMMERCE'
            ],
            "merchant" => [
                "mid" => $this->module->getMerchantMid(),
                "tid" => $this->module->getMerchantTid(),
            ],
            "customer" => [
                "firstName" => $customer->firstname,
                "lastName" => $customer->lastname,
                "birthDate" => $customer->birthday === "0000-00-00" ? null : $customer->birthday,
                "contact" => [
                    "email" => $customer->email,
                    "mobile" => $address->phone_mobile,
                    "phone" => $address->phone,
                    "ip" => $this->module->getClientIpAddress(),
                ],
                "billingAddress" => [
                    "street1" => $address->address1,
                    "street2" => $address->address2,
                    "city" => $address->city,
                    "state" => $state->name,
                    "postCodeZip" => $address->postcode,
                    "countryAlphaCode" => $country->iso_code,
                ]
            ],
            "additionalData" => []
        ];

        return $this->sendRequest('POST', 'payments/reversal', $data);
    }

    /**
     * Sends the cURL request to the API.
     *
     * @param string $method POST | GET
     * @param string $endpoint The endpoint to send to.
     * @param array $data The values to send to the API
     * @return string
     */
    private function sendRequest($method, $endpoint, $data)
    {
        $headers = array(
            "Authorization: Bearer " . $this->module->getAuthToken(), 
            "Accept: application/json", 
            "Content-Type: application/json; charset=utf-8"
        );

        $queryParams = $method === "GET" ? http_build_query($data) : null;
        $url = $this->module->getFullTruevoEndpoint($endpoint, $queryParams);

        $this->module->truevoLogger('Truevo payment sending cURL request. Check log file for more details.');
        $this->module->truevoFileLogger('Truevo payment sending cURL request to ' . $url . ' with the following data: ' . print_r($data, true));

        $this->module->_serializePrecisionFix();
        
        $curl = curl_init();

        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_ENCODING, "");
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, 30);
        curl_setopt($curl, CURLOPT_TIMEOUT, 30);

        if ($method === "POST") {
            $headers[] = "IdempotencyKey: " . $this->idempotencyKey;

            curl_setopt($curl, CURLOPT_POST, 1);
            curl_setopt($curl, CURLOPT_POSTFIELDS, json_encode($data));
        }

        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);

        $response = curl_exec($curl);

        if (curl_errno($curl)) {
            $this->module->truevoLogger('Error Sending Request to Truevo API. Check log file for more details.');
            $this->module->truevoFileLogger('Truevo payment error sending cURL request: ' . print_r(curl_error($curl), true));
            return false;
        }

        curl_close($curl);

        return json_decode($response);
    }
}
