<?php

require_once dirname(__FILE__) . '/BaseTruevoAdminController.php';

class AdminRefundOrderController extends BaseTruevoAdminController
{
    /**
     * The max amount allowed to be refunded.
     *
     * @var float
     */
    private $availableForRefund;

    /**
     * The total products in the order.
     *
     * @var integer
     */
    private $totalProducts;

    /**
     * The input variables sent back from the form.
     *
     * @var array
     */
    private $_input = array();

    /**
     * The errors in the form.
     *
     * @var array
     */
    private $_errors = array();

    /**
     * Constructor
     */
    public function __construct()
    {
        parent::__construct();

        $this->availableForRefund   = $this->getMaxRefundableAmount();
        $this->totalProducts        = $this->getTotalProducts();

        // Prestashop specific
        $this->meta_title           = $this->generateMetaTitle();
        $this->toolbar_title        = $this->generatePageTitle();
        $this->bootstrap            = true;
    }

    /**
     * Assign smarty variables for all default views, list and form, then call other init functions.
     */
    public function initContent()
    {
        parent::initContent();

        $this->context->smarty->assign([
            'currency_iso_code'     => $this->getCurrencyIsoCode(),
            'min_refund_amount'     => 0,
            'max_refund_amount'     => $this->availableForRefund,
            'min_number_of_pieces'  => 0,
            'max_number_of_pieces'  => $this->totalProducts,
        ]);

        $this->setTemplate('form.tpl');
    }

    /**
     * Send through the refund details to the Truevo API.
     */
    public function postProcess()
    {
        if (Tools::isSubmit('truevo_refund_form')) {
            $this->_input = Tools::getAllValues();

            $this->validate();

            if (count($this->_errors) === 0) {
                $response = $this->truevoApi->refund(
                    (float) $this->_input['refund_amount'], 
                    (int) $this->_input['number_of_pieces'], 
                    $this->_input['reason_for_refund']
                );

                // Responses
                if ($response) {
                    if ($response->outcome->status === "RECEIVED") {
                        $transactionUniqueIdentifier = $response->transactionUniqueIdentifier;

                        $log = $this->buildLogArray('REVERSAL', $response, ['transaction_unique_identifier' => $transactionUniqueIdentifier]);
                        $this->logToOrder($this->order, $log);

                        switch ( $this->pollingTransactionStatus($transactionUniqueIdentifier) ) {
                            case 'COMPLETE':
                                if ((float) $this->_input['refund_amount'] === $this->availableForRefund) {
                                    $this->order->setCurrentState(Configuration::get('PS_OS_CANCELED'));
                                }

                                $this->assignStatusNotification('success', 'Truevo refund request successfully sent.');
                                break;
                            case 'DECLINED':
                                $this->assignStatusNotification('danger', 'Truevo payment declined the refund.');
                                break;
                            default: 
                                // Refund is still pending
                                $this->assignStatusNotification('info', 'Refund is still pending.');
                                break;
                        }

                        $this->redirectToOrderView($this->order->id);

                    }
                }

                // Error capturing payment
                $this->assignStatusNotification('danger', 'There was an error requesting a refund for the Truevo payment.');
                $this->redirectToOrderView($this->order->id);
            }
            
            // Redirect back with validation errors
            $this->context->smarty->assign([
                'input'  => $this->_input,
                'errors' => $this->_errors
            ]);
        }
    }

    /**
     * Validate the refund form details sent through and return any errors.
     *
     * @return array
     */
    private function validate()
    {
        $errors = array();

        $details = $this->truevoApi->getDetails();
        $this->availableForRefund = $details->paymentDetails->availableForRefund;

        $refundAmount   = (float) $this->_input['refund_amount'];
        $numberOfPieces = (int) $this->_input['number_of_pieces'];
        $refundReason   = $this->_input['reason_for_refund'];

        #region Refund Amount
        if ($refundAmount != number_format($refundAmount, 2))
            $errors['refund_amount'][] = 'Please enter a valid amount.';

        if ($refundAmount < 0) 
            $errors['refund_amount'][] = 'Negative values are not allowed for the refund amount.';

        if ($refundAmount > $this->availableForRefund)
            $errors['refund_amount'][] = 'Please enter an amount that does not exceed the maximum available refund amount.';
        #endregion

        #region Number of Pieces
        if ($numberOfPieces < 0) 
            $errors['number_of_pieces'][] = 'Negative values are not allowed for the number of pieces.';

        if ($numberOfPieces > $this->getTotalProducts())
            $errors['number_of_pieces'][] = 'Please enter an amount that does not exceed the maximum number of products in the order.';
        #endregion

        #region Refund reason
        if (strlen($refundReason) > 150) 
            $errors['reason_for_refund'][] = 'Please make sure your reason for refund does not exceed 150 characters.';
        #endregion

        // Assign the errors to a class specific var
        $this->_errors = $errors;
    }

    /**
     * Generate the meta title for the page.
     *
     * @return string
     */
    private function generateMetaTitle()
    {
        $customer = new Customer($this->order->id_customer);
        return sprintf(
            'Refund & Reversal > Order %s from %s %s',
            $this->order->reference,
            $customer->firstname,
            $customer->lastname
        );
    }

    /**
     * Set breadcrumbs array for the controller page.
     *
     * @param int|null $tab_id
     * @param array|null $tabs
     */
    public function initBreadcrumbs($tab_id = null, $tabs = null)
    {
        $breadcrumbs = [
            'container' => ['name' => 'Orders', 'href' => '', 'icon' => ''],
            'tab'       => ['name' => 'Refund', 'href' => '', 'icon' => ''],
        ];

        $this->context->smarty->assign([
            'breadcrumbs2' => $breadcrumbs,
            'quick_access_current_link_name' => Tools::safeOutput($breadcrumbs['tab']['name']),
            'quick_access_current_link_icon' => $breadcrumbs['container']['icon'],
        ]);
    }

    /**
     * Generate the page title.
     *
     * @return string
     */
    private function generatePageTitle()
    {
        $customer = new Customer($this->order->id_customer);
        return sprintf(
            'Refund Order #%s - %s from %s %s',
            $this->order->id,
            $this->order->reference,
            $customer->firstname,
            $customer->lastname
        );
    }

    /**
     * Get the total number of products being ordered.
     *
     * @return integer
     */
    private function getTotalProducts()
    {
        $cart = new Cart($this->order->id_cart);

        $quantity = 0;
        foreach ($cart->getProducts() as $product) {
            $quantity += (int) $product['cart_quantity'];
        }

        return $quantity;
    }

    /**
     * Get the maximum amount refundable via the params.
     *
     * @return float
     */
    private function getMaxRefundableAmount()
    {
        return (float) Tools::getValue('available_for_refund');
    }
}
