<?php

require_once dirname(__FILE__) . '/../../classes/TruevoAPI.php';

use Symfony\Component\Routing\Exception\RouteNotFoundException;

class BaseTruevoAdminController extends ModuleAdminController
{
    /**
     * How many times to attempt to fetch a transaction status before aborting.
     */
    const MAX_QUERY_STATUS_ATTEMPTS = 5;

    /**
     * How many attempts to get the transaction status have already been made.
     *
     * @var integer
     */
    private $attempts = 0;

    /**
     * The order being refunded or reversed.
     *
     * @var Order
     */
    protected $order;

    /**
     * The Truevo API handler class.
     *
     * @var TruevoAPI
     */
    protected $truevoApi;

    /**
     * Constructor
     */
    public function __construct()
    {
        $this->order     = new Order((int) Tools::getValue('id_order'));
        $this->truevoApi = new TruevoAPI($this->order);

        parent::__construct();
    }

    /**
     * Assign response message to the cookie to display on the order page.
     * Gets used in truevopayments.php > displayCaptureOfPurchaseNotification()
     *
     * @param string $type "success" | "info" | "warning" | "danger"
     * @param string $message
     * 
     * @return void
     */
    protected function assignStatusNotification(string $type = "success", string $message = "Successfully completed!")
    {
        $orderId = $this->order->id;

        $this->module->assignStatusNotification($orderId, $type, $message);
    }

    /**
     * Get the url of the show order page via the order ID.
     *
     * @param integer $orderId
     * @return string
     */
    protected function redirectToOrderView(int $orderId) 
    {
        try {
            $url = $this->context->link->getAdminLink('AdminOrders', true, array('route' => 'admin_orders_view', 'orderId' => $orderId));
        } catch (RouteNotFoundException $exception) {
            $url = $this->context->link->getAdminLink('AdminOrders', true, [], ['id_order' => $orderId]) . '&vieworder';
        }

        Tools::redirectAdmin($url);
    }

    /**
     * Get the ISO code of the currency used in this order.
     *
     * @return string
     */
    protected function getCurrencyIsoCode()
    {
        $currency = new Currency($this->order->id_currency);

        return $currency->iso_code;
    }

    /**
     * Build an array from the outcome of a response.
     *
     * @param string $action CAPTURE | QUERY STATUS | REFUND | REVERSAL
     * @param object $response
     * @param array $extraData
     * @return array
     */
    protected function buildLogArray($action, $response, $extraData = [])
    {
        // Rebuild the additional data to a proper key value pair array
        $additionalData = array();
        foreach ($response->additionalData as $keyValuePair) {
            $additionalData[$keyValuePair->key] = $keyValuePair->value;
        }
        
        $data = [
            'action'             => $action,
            'status'             => $response->outcome->status,
            'transaction_status' => $response->outcome->transactionStatus,
            'response_code'      => $response->outcome->responseCode ?? null,
            'reason_description' => $response->outcome->reasonDescription ?? null,
            'amount'             => isset($response->amount->totalAmount) ? $this->getCurrencyIsoCode() . number_format($response->amount->totalAmount, 2) : null,
            'number_of_pieces'   => $additionalData['number_of_pieces'] ?? null,
            'reason_for_refund'  => $additionalData['reason_for_refund'] ?? null
        ];

        $return = array_merge($data, $extraData);

        return array_filter($return);
    }

    /**
     * Add a message to the customer thread of an order.
     *
     * @param Order $order
     * @param array|string $content
     * @return void
     */
    protected function logToOrder(Order $order, $content)
    {
        if (is_array($content)) {
            $message = '';
            foreach ($content as $key => $value) {
                $name = str_replace('_', ' ', $key);
                $message .= ucwords($name) . ": " . $value . "\n";
            }
        } else if (is_string($content)) {
            $message = $content;
        } else {
            return;
        }

        $customer = new Customer($order->id_customer);

        $customer_thread = new CustomerThread();
        $customer_thread->id_contact = 0;
        $customer_thread->id_customer = (int) $order->id_customer;
        $customer_thread->id_shop = (int) $this->context->shop->id;
        $customer_thread->id_order = (int) $order->id;
        $customer_thread->id_lang = (int) $this->context->language->id;
        $customer_thread->email = $customer->email;
        $customer_thread->status = 'closed';
        $customer_thread->token = Tools::passwdGen(12);
        $customer_thread->add();

        $customer_message = new CustomerMessage();
        $customer_message->id_customer_thread = $customer_thread->id;
        $customer_message->id_employee = 0;
        $customer_message->message = $message;
        $customer_message->private = 1;

        if (!$customer_message->add()) {
            PrestaShopLogger::addLog('Truevo payment error. Failed to capture response from cURL request.', 3, null, 'Order', (int) $order->id);
        }
    }

    /**
     * Poll the status for a transaction unique identifier until either a 
     * "COMPLETE" or "DECLINED" is returned or the request times out.
     *
     * @param string $transactionId
     * @return string "COMPLETE", "DECLINED", or "PENDING"
     */
    protected function pollingTransactionStatus($transactionId)
    {
        $status = "PENDING";
        $isComplete = false;

        do {
            sleep(1); // Wait a second before querying the status
            $this->attempts += 1; // Increment the total number of attempts

            $response = $this->truevoApi->getStatus($transactionId);

            if ($response) {
                $status = $response->outcome->transactionStatus;

                // Don't log pending requests unless it is the final one
                if ($status !== "PENDING" || $this->attempts === self::MAX_QUERY_STATUS_ATTEMPTS) {
                    $log = $this->buildLogArray('QUERY STATUS', $response);
                    $this->logToOrder($this->order, $log);
                }
                
                $isComplete = $status !== "PENDING"; 
            }
        } while ($this->attempts < self::MAX_QUERY_STATUS_ATTEMPTS && !$isComplete);

        return $status;
    }
}