<?php

class TruevoPaymentsPaymentWidgetModuleFrontController extends ModuleFrontController
{
    public function initContent()
    {
        parent::initContent();

        $cart = $this->context->cart;

        if ($cart->id_customer == 0 || $cart->id_address_delivery == 0 || $cart->id_address_invoice == 0 || !$this->module->active) {
            Tools::redirect($this->module->getOrderUrlPath());
        }

        // Check that this payment option is still available in case the customer changed his address just before the end of the checkout process
        $authorized = false;
        foreach (Module::getPaymentModules() as $module) {
            if ($module['name'] == 'truevopayments') {
                $authorized = true;
                break;
            }
        }

        if (!$authorized) {
            die($this->trans('This payment method is not available.', array(), 'Modules.Truevopayments.Shop'));
        }

        $customer = new Customer($cart->id_customer);

        if (!Validate::isLoadedObject($customer)) {
            Tools::redirect($this->module->getOrderUrlPath());
        }

        $cancelUrl = $this->module->getOrderUrlPath();
        parse_str( parse_url($cancelUrl, PHP_URL_QUERY), $cancelUrlQueryParams );

        $this->module->_serializePrecisionFix();

        $this->context->smarty->assign([
            'jsSrc' => $this->module->getTruevoWidgetUrl(),
            'configuration' => json_encode($this->prepareWidgetConfiguration()),
            'callbackUrl' => $this->context->link->getModuleLink('truevopayments', 'callback', array(), true),
            'cancelUrl' => strtok($cancelUrl, '?'),
            'cancelUrlQueryParams' => $cancelUrlQueryParams
        ]);

        $this->setTemplate('module:truevopayments/views/templates/front/payment_widget.tpl');
    }

    public function setMedia()
    {
        parent::setMedia();

        $this->registerStylesheet(
            'truevopayments-style',
            'modules/'.$this->module->name.'/css/truevopayments.css',
            [
                'media' => 'all',
                'priority' => 200,
            ]
        );
    }

    protected function prepareArgs()
    {
        $cart = $this->context->cart;

        $shop     = new Shop($cart->id_shop);
        $currency = new Currency($cart->id_currency);
        $customer = new Customer($cart->id_customer);
        $address  = new Address($cart->id_address_invoice);
        $state    = new State($address->id_state);
        $country  = new Country($address->id_country);

        $args = [
            'transaction_type'      => $this->module->getTransactionType(),
            'merchant_id'           => $this->module->getMerchantMid(),
            'terminal_id'           => $this->module->getMerchantTid(),
            'language'              => $this->module->getTruevoLocale(),
            'show_logo'        	    => $this->module->showTruevoLogo(),
            'widget_theme'          => $this->module->getWidgetTheme(),
            'css_styles'            => $this->minifyCss($this->module->getCustomCss()),
            'payment_reference'     => $shop->name . ' Purchase',
            'order_reference'       => '#' . $cart->id . ' | ' . date( DATE_W3C ),
            'currency'       	    => $currency->iso_code,
            'amount'                => (float) $cart->getOrderTotal(true, Cart::BOTH),
            'gratuity_amount'       => 0,
            'transaction_source'    => 'ECOMMERCE',
            'customer_first_name'   => $customer->firstname,
            'customer_last_name'    => $customer->lastname,
            'customer_birth_date'   => $customer->birthday === "0000-00-00" ? null : $customer->birthday,
            'customer_email'        => $customer->email,
            'customer_mobile'       => $address->phone_mobile,
            'customer_phone'        => $address->phone,
            'billing_address'       => $address->address1,
            'billing_address_2'     => $address->address2,
            'billing_city'          => $address->city,
            'billing_state'         => $state->name,
            'billing_country'       => $country->iso_code,
            'billing_postcode'      => $address->postcode,
            'ip_address'            => $this->module->getClientIpAddress(),
            'additional_data'       => []
        ];

        // If tokenization is enabled and customer is logged in append widget configuration
        if ($this->module->isTokenizationEnabled() && $this->context->customer->isLogged()) {
            $args['tokenization_shopper_reference']     = $this->module->getCustomerTokenizationReference($customer);
            $args['tokenization_registration_process']  = 'Optional';
            $args['tokenization_type']                  = 'OneClickPayment';
        }

        return $args;
    }

    protected function prepareWidgetConfiguration()
    {
        $args = $this->prepareArgs();

        $this->module->truevoLogger('Truevo payment request prepared. Check log file for more details.');
        $this->module->truevoFileLogger('Truevo payment request prepared. ' . print_r($args, true));

        $configuration = [
            "transactionType" => $args['transaction_type'],
            "widgetConfiguration" => [
                "language" => $args['language'],
                "showTruevo" => $args['show_logo'],
                "iframeStyles" => [
                    "theme" => $args['widget_theme'],
                    "style" => $args['css_styles'],
                ]
            ],
            "transaction" => [
                "paymentReference" => $args['payment_reference'],
                "merchantReference" => $args['order_reference'],
                "amount" => [
                    "totalAmount" => $args['amount'],
                    "currencyAlphaCode" => $args['currency'],
                    "details" => [
                        "paymentAmount" => $args['amount'],
                        "gratuityAmount" => $args['gratuity_amount']
                    ],
                ],
                "source" => $args['transaction_source']
            ],
            "merchant" => [
                "mid" => $args['merchant_id'],
                "tid" => $args['terminal_id'],
            ],
            "customer" => [
                "firstName" => $args['customer_first_name'],
                "lastName" => $args['customer_last_name'],
                "birthDate" => $args['customer_birth_date'],
                "contact" => [
                    "email" => $args['customer_email'],
                    "mobile" => $args['customer_mobile'],
                    "phone" => $args['customer_phone'],
                    "ip" => $args['ip_address'],
                ],
                "billingAddress" => [
                    "street1" => $args['billing_address'],
                    "street2" => $args['billing_address_2'],
                    "city" => $args['billing_city'],
                    "state" => $args['billing_state'],
                    "postCodeZip" => $args['billing_postcode'],
                    "countryAlphaCode" => $args['billing_country'],
                ]
            ],
            "additionalData" => $args['additional_data']
        ];

        // If tokenization attributes exist, pass them through to the configuration
        if (isset($args['tokenization_shopper_reference']) && isset($args['tokenization_registration_process']) && isset($args['tokenization_type'])) {
            $configuration['tokenization'] = [
                "merchantShopperReference" => $args['tokenization_shopper_reference'],
                "tokenRegistrationProcess" => $args['tokenization_registration_process'],
                "tokenType" => $args['tokenization_type']
            ];
        }

        return $configuration;
    }

    protected function minifyCss($css)
    {
        return str_replace(array("\r", "\n"), '', $css);
    }

    static public function getIsoByName($country_name) {
        $sql='
        SELECT `id_country`
        FROM `'._DB_PREFIX_.'country_lang`
        WHERE `name` = "'.$country_name.'"';
        $result = Db::getInstance()->getRow($sql);
        $iso=Country::getIsoById($result['id_country']);
        return $iso;
    }
}
