<?php

use PrestaShop\PrestaShop\Core\Payment\PaymentOption;
use PrestaShopBundle\Controller\Admin\Sell\Order\ActionsBarButton;

if (!defined('_PS_VERSION_')) {
    exit;
}

require_once dirname(__FILE__) . '/classes/TruevoAPI.php';

class TruevoPayments extends PaymentModule
{
    const MODULE_NAME = 'truevopayments';

    const ON_HOLD_STATE = 'On Hold';

    private $_truevoTabs = ['AdminProcessOrder', 'AdminRefundOrder', 'AdminReverseOrder'];

    protected $_html = '';
    protected $_postErrors = array();

    public function __construct()
    {
        $this->name = self::MODULE_NAME;
        $this->tab = 'payments_gateways';
        $this->version = '1.4.1';
        $this->ps_versions_compliancy = array('min' => '1.7.1.0', 'max' => _PS_VERSION_);
        $this->author = 'Truevo';
        $this->controllers = array('truevopayments', 'callback');
        $this->is_eu_compatible = 1;

        $this->currencies = true;
        $this->currencies_mode = 'checkbox';

        $this->bootstrap = true;
        parent::__construct();

        $this->displayName = $this->trans('Truevo Payment Gateway', array(), 'Modules.Truevopayments.Admin');
        $this->description = $this->trans(' Take online payments with the Truevo Payment Plugin. Tokenisation lets your customers save their card details for repeat purchase convenience. Perform reversals, refunds and captures. PCI compliant and 3DS v2 enabled, to keep you and your customers safe.', array(), 'Modules.Truevopayments.Admin');
        
        if (!count(Currency::checkPaymentCurrencies($this->id))) {
            $this->warning = $this->l('No currency has been set for this module.');
        }
    }

    public function install()
    {
        return parent::install()
            && $this->registerHook('paymentOptions')
            && $this->registerHook('actionValidateOrder')
            && $this->registerHook('displayAdminOrder')
            && (version_compare(_PS_VERSION_, '1.7.7', '>=') ? $this->registerHook('actionGetAdminOrderButtons') : $this->registerHook('displayBackOfficeOrderActions'))
            && (version_compare(_PS_VERSION_, '1.7.7', '>=') ? $this->registerHook('displayAdminOrderTop') : $this->registerHook('displayInvoice'))
            && $this->createResponseDetailsTable()
            && $this->createAdditionalOrderStates();
    }

    public function uninstall()
    {
        return parent::uninstall()
            && $this->unregisterHook('paymentOptions')
            && $this->unregisterHook('actionValidateOrder')
            && $this->unregisterHook('displayAdminOrder')
            && (version_compare(_PS_VERSION_, '1.7.7', '>=') ? $this->unregisterHook('actionGetAdminOrderButtons') : $this->unregisterHook('displayBackOfficeOrderActions'))
            && (version_compare(_PS_VERSION_, '1.7.7', '>=') ? $this->unregisterHook('displayAdminOrderTop') : $this->unregisterHook('displayInvoice'))
            && $this->removeAdditionalOrderStates();
    }

    public function createResponseDetailsTable()
    {
        $sql = "CREATE TABLE IF NOT EXISTS " ._DB_PREFIX_ . "truevo_payment_response_details (
            `id_response_detail` INT(10) NOT NULL AUTO_INCREMENT PRIMARY KEY,
            `id_cart` INT(10) NOT NULL,
            `id_order` INT(10),
            `transaction_type` VARCHAR(20),
            `rrn` VARCHAR(50),
            `transaction_id` VARCHAR(50),
            `transaction_status` VARCHAR(20),
            `response_code` VARCHAR(10),
            `merchant_reference` VARCHAR(50),
            `payment_reference` VARCHAR(50))";

        return Db::getInstance()->execute($sql);
    }

    public function installTab()
    {
        $result = false;

        foreach ($this->_truevoTabs as $className) {
            $tab = new Tab();
            $tab->class_name = $className;
            $tab->module = $this->name;
            $tab->active = true;
            $tab->id_parent = -1;
            $tab->name = array_fill_keys(
                Language::getIDs(false),
                $this->displayName
            );

            $result &= $tab->add();
        }
        
        return $result;
    }

    public function uninstallTab()
    {
        $result = false;

        foreach ($this->_truevoTabs as $className) {
            $id_tab = (int) Tab::getIdFromClassName($className);
            $tab = new Tab($id_tab);

            if (Validate::isLoadedObject($tab)) {
                $result &= $tab->delete();
            }
        }
        
        return $result;
    }

    public function createAdditionalOrderStates()
    {
        $states = OrderState::getOrderStates((int)$this->context->language->id, false);

        // Check if order state exist
        $existingId = null;
        foreach ($states as $state) {
            if ($state['module_name'] === $this->name && $state['name'] === self::ON_HOLD_STATE) {
                $existingId = $state['id_order_state'];
                break;
            }
        }

        // On Hold 
        $order_state              = new OrderState();
        $order_state->id          = $existingId;
        $order_state->color       = '#6C868E';
        $order_state->send_email  = false;
        $order_state->module_name = $this->name;
        $order_state->name        = array();
        $order_state->deleted     = false;
        $languages                = Language::getLanguages(false);
        foreach ($languages as $language) {
            $order_state->name[$language['id_lang']] = self::ON_HOLD_STATE;
        }

        $order_state->save();

        $result = Configuration::updateValue('TRUEVO_ON_HOLD', $order_state->id);

        return $result;
    }

    public function removeAdditionalOrderStates()
    {
        return Db::getInstance()->update('order_state', array(
            'deleted' => 1
        ), "`module_name` = '" . $this->name . "'");
    }
    
    public function _processConfiguration($_config) 
    {
        $conf = $_config;

        $_formFields = array(
            'form_lang' => array(
                'EN'  => 'English',
            ),
            'widget_themes' => array(
                'plain' => 'Plain',
                'card' => 'Card'
            ),
            'debug_log_path' => $this->getDebugLogFilePath()
        );

        foreach($_formFields as $key => $value) {
            $this->context->smarty->assign($key, $value);
        }

        if (Tools::isSubmit('truevo_pc_form')) {
            $this->context->smarty->assign('confirmation', 'ok');
            
            foreach ($conf as $value) {
                Configuration::updateValue($value, Tools::getValue($value));
            }
        }
    }

    public function _assignConfiguration($_config) 
    {
        $conf = $_config;

        foreach ($conf as $value) {
            $this->context->smarty->assign($value, Configuration::get($value));
        }
    }

    public function getContent() 
    {
        $_config = array(
            'TRUEVO_ENABLE_PAYMENT_OPTION',
            'TRUEVO_TITLE',
            'TRUEVO_DESC',
            'TRUEVO_ENABLE_PREAUTH',
            'TRUEVO_ENABLE_TOKENIZATION',
            'TRUEVO_TEST_AUTH_TOKEN',
            'TRUEVO_PROD_AUTH_TOKEN',
            'TRUEVO_WIDGET_THEME',
            'TRUEVO_WIDGET_LANGUAGE',
            'TRUEVO_SHOW_LOGO',
            'TRUEVO_TEST_MODE',
            'TRUEVO_TEST_MID',
            'TRUEVO_TEST_TID',
            'TRUEVO_PROD_MID',
            'TRUEVO_PROD_TID',
            'TRUEVO_CSS_STYLES',
            'TRUEVO_DEBUG_LOGGING'
        );

        $this->_processConfiguration($_config);
        $this->_assignConfiguration($_config);

        return $this->display(__FILE__, 'getContent.tpl');
    }

    public function hookPaymentOptions($params)
    {
        if (!$this->active) {
            return;
        }

        if (!$this->checkCurrency($params['cart'])) {
            return;
        }

        if ($this->isPaymentMethodEnabled()) {
            $payment_options = [
                $this->getWidgetPaymentOption(),
            ];
        } else {
            $payment_options = [];
        }
        
        return $payment_options;
    }

    public function checkCurrency($cart)
    {
        $currency_order = new Currency($cart->id_currency);
        $currencies_module = $this->getCurrency($cart->id_currency);

        if (is_array($currencies_module)) {
            foreach ($currencies_module as $currency_module) {
                if ($currency_order->id == $currency_module['id_currency']) {
                    return true;
                }
            }
        }

        return false;
    }

    public function hookActionValidateOrder($params)
    {		
        $order = $params["order"];
        $customer = $params["customer"];

        $result = $this->getTransactionDetailsFromCartId((int) $order->id_cart);

        if ($result) {
            $message = "Transaction Type: " . $result['transaction_type'] . "\n";
            $message .= "RRN: " . $result['rrn'] . "\n";
            $message .= "Transaction ID: " . $result['transaction_id'] . "\n";
            $message .= "Transaction Status: " . $result['transaction_status'] . "\n";
            $message .= "Response Code: " . $result['response_code'] . "\n";
            $message .= "Merchant Reference: " . $result['merchant_reference'];
        }

        if (isset($message)) {
            $customer_thread = new CustomerThread();
            $customer_thread->id_contact = 0;
            $customer_thread->id_customer = (int) $order->id_customer;
            $customer_thread->id_shop = (int) $this->context->shop->id;
            $customer_thread->id_order = (int) $order->id;
            $customer_thread->id_lang = (int) $this->context->language->id;
            $customer_thread->email = $customer->email;
            $customer_thread->status = 'closed';
            $customer_thread->token = Tools::passwdGen(12);
            $customer_thread->add();

            $customer_message = new CustomerMessage();
            $customer_message->id_customer_thread = $customer_thread->id;
            $customer_message->id_employee = 0;
            $customer_message->message = $message;
            $customer_message->private = 1;

            if (!$customer_message->add()) {
                PrestaShopLogger::addLog('Truevo payment error. Failed to capture transaction details.', 3, null, 'Order', (int) $order->id);
            }

            $this->updateTransactionDetailsWithOrderId($result['id_response_detail'], $order->id);
        }

        // A regular "SALE" already creates a payment history
        if ($result['transaction_type'] === 'AUTHORIZE') {
            $order->addOrderPayment(
                0,
                $this->displayName,
                $result['transaction_id'],
                new Currency($order->id_currency)
            );
        }
    }

    /**
     * Add buttons to main buttons bar (Prestashop Version >= 1.7.7)
     */
    public function hookActionGetAdminOrderButtons(array $params)
    {
        $order = new Order($params['id_order']);

        // Check if this payment method was used
        if ($order->module === self::MODULE_NAME) {
            // Truevo API
            $truevoApi = new TruevoAPI($order);
            $response = $truevoApi->getDetails();

            if ($response) {
                /** @var \PrestaShopBundle\Controller\Admin\Sell\Order\ActionsBarButtonsCollection $bar */
                $bar = $params['actions_bar_buttons_collection'];

                // Only show the capture button if it is allowed and the order status is On Hold
                if ($response->paymentDetails->isCaptureAllowed && $order->getCurrentState() === Configuration::get('TRUEVO_ON_HOLD')) {
                    $bar->add(
                        new ActionsBarButton(
                            'btn-primary', 
                            ['href' => $this->context->link->getAdminLink('AdminProcessOrder', true, [], ['id_order' => (int) $order->id])], 
                            'Capture Pre-authorisation'
                        )
                    );
                }

                // Only show refund button if it is allowed
                if ($response->paymentDetails->isRefundAllowed && $response->paymentDetails->availableForRefund > 0) {
                    $bar->add(
                        new ActionsBarButton(
                            'btn-secondary', 
                            ['href' => $this->context->link->getAdminLink('AdminRefundOrder', true, [], ['id_order' => (int) $order->id, 'available_for_refund' => $response->paymentDetails->availableForRefund])], 
                            'Refund order'
                        )
                    );
                }

                // Only show reversal button if it is allowed
                if ($response->paymentDetails->isReverseAllowed) {
                    $bar->add(
                        new ActionsBarButton(
                            'btn-secondary', 
                            ['href' => $this->context->link->getAdminLink('AdminReverseOrder', true, [], ['id_order' => (int) $order->id])], 
                            'Reverse order'
                        )
                    );
                }
            }
        }
    }

    /**
     * Add buttons to main buttons bar (Prestashop Version < 1.7.7)
     */
    public function hookDisplayBackOfficeOrderActions(array $params)
    {        
        $order = new Order($params['id_order']);

        // Check if this payment method was used
        if ($order->module === self::MODULE_NAME) {
            // Truevo API
            $truevoApi = new TruevoAPI($order);
            $response = $truevoApi->getDetails();

            if ($response) {
                $this->context->smarty->assign([
                    'orderId'           => (int) $order->id,
                    'isCaptureAllowed'  => ($response->paymentDetails->isCaptureAllowed && $order->getCurrentState() === Configuration::get('TRUEVO_ON_HOLD')),
                    'isRefundAllowed'   => ($response->paymentDetails->isRefundAllowed && $response->paymentDetails->availableForRefund > 0),
                    'isReverseAllowed'  => $response->paymentDetails->isReverseAllowed,
                    'processUrl'        => $this->context->link->getAdminLink('AdminProcessOrder'),
                    'refundUrl'         => $this->context->link->getAdminLink('AdminRefundOrder', true, [], ['id_order' => (int) $order->id, 'available_for_refund' => $response->paymentDetails->availableForRefund]),
                    'reverseUrl'        => $this->context->link->getAdminLink('AdminReverseOrder', true, [], ['id_order' => (int) $order->id]),
                ]);
            }

            return $this->fetch('module:truevopayments/views/templates/hook/displayBackOfficeOrderActions.tpl');
        }
    }

    /**
     * This hook displays content at the top of the order view page (Prestashop Version >= 1.7.7)
     */
    public function hookDisplayAdminOrderTop(array $params)
    {
        return $this->displayCaptureOfPurchaseNotification($params['id_order']);
    }

    /**
     * This hook displays new blocks on the invoice (order) (Prestashop Version < 1.7.7)
     */
    public function hookDisplayInvoice(array $params)
    {
        return $this->displayCaptureOfPurchaseNotification($params['id_order']);
    }

    /**
     * Get the response from the Truevo API and display it as a notification on the order page.
     * 
     * @var int $orderId
     */
    private function displayCaptureOfPurchaseNotification($orderId)
    {
        $cookieKey = "truevo_api_notification_order_$orderId";

        $truevoCaptureNotification = json_decode($this->context->cookie->__get($cookieKey));

        $this->context->cookie->__unset($cookieKey);

        if ($truevoCaptureNotification && time() <= $truevoCaptureNotification->expiry_time) {
            $this->context->smarty->assign('truevoCaptureNotification', [
                'type'    => $truevoCaptureNotification->type,
                'message' => $truevoCaptureNotification->message,
            ]);

            return $this->fetch("module:truevopayments/views/templates/hook/displayCaptureOfPurchaseNotification.tpl");
        }
    }

    /**
     * Display new elements in the Back Office, tab AdminOrder.
     * This hook launches modules when the AdminOrder tab is displayed in the Back Office.
     */
    public function hookDisplayAdminOrder(array $params)
    {
        $order = new Order($params['id_order']);
        
        // Check if this payment method was used
        if ($order->module === self::MODULE_NAME) {
            // This is where the partial refund button gets hidden
            return $this->fetch('module:truevopayments/views/templates/hook/displayAdminOrder.tpl');
        }
    }

    public function getWidgetPaymentOption()
    {
        $this->context->smarty->assign([
            'truevoDesc' => $this->getPaymentMethodDescription()
        ]);
        
        $widgetOption = new PaymentOption();
        $widgetOption->setCallToActionText($this->getPaymentMethodTitle())
                        ->setAction($this->context->link->getModuleLink($this->name, 'paymentwidget', array(), true))
                        ->setAdditionalInformation($this->context->smarty->fetch('module:truevopayments/views/templates/front/payment_description.tpl'))
                        ->setLogo(Media::getMediaPath(_PS_MODULE_DIR_ . $this->name . '/payment.jpg'));

        return $widgetOption;
    }

    public function getOrderUrlPath($step = 1)
    {
        return $this->context->link->getPageLink('order', true, null, array('step' => $step));
    }

    public function getTransactionDetailsFromCartId(int $cartId) 
    {
        return Db::getInstance()->getRow("SELECT * FROM `" . _DB_PREFIX_ . "truevo_payment_response_details` WHERE `id_cart` = '" . (int) $cartId  . "';");
    }

    public function getTransactionDetailsForOrder(Order $order) 
    {
        $result = Db::getInstance()->getRow("SELECT * FROM `" . _DB_PREFIX_ . "truevo_payment_response_details` WHERE `id_order` = '" . (int) $order->id  . "';");

        if ($result['transaction_type'] === "AUTHORIZE") {
            $captureTransactionId = $this->getCaptureOrderPayment($order);

            if ($captureTransactionId) {
                $result['original_transaction_id'] = $result['transaction_id'];
                $result['transaction_id'] = $captureTransactionId;
            }
        }

        return $result;
    }

    public function getCaptureOrderPayment(Order $order)
    {
        $transactionId = null;

        $payments = $order->getOrderPayments();

        foreach ($payments as $payment) {
            if (strtolower($payment->payment_method) === strtolower($this->displayName . ' (Capture)')) {
                $transactionId = $payment->transaction_id;
                break;
            }
        }

        return $transactionId;
    }

    public function updateTransactionDetailsWithOrderId(int $responseDetailsId, int $orderId)
    {
        return Db::getInstance()->update('truevo_payment_response_details', array(
            'id_order' => (int) $orderId,
        ), 'id_response_detail = ' . (int) $responseDetailsId, 1);
    }

    public function getClientIpAddress()
    {
        $ipAddress = isset($_SERVER['REMOTE_ADDR']) ? trim($_SERVER['REMOTE_ADDR']) : null;

        // Check IP address is valid and not localhost
        if (filter_var($ipAddress, FILTER_VALIDATE_IP) && (substr($ipAddress, 0, 4) != '127.' && $ipAddress != '::1')) {
            return $ipAddress;
        }

        return null;
    }

    public function getCustomerTokenizationReference(Customer $customer)
    {
        return "tokenization-" . $customer->id . '-' . $customer->email . '-' . $customer->secure_key;
    }

    public function truevoLogger($message)
    {
        if (Configuration::get('TRUEVO_DEBUG_LOGGING')) {
            PrestaShopLogger::addLog($message, 1);
        }
    }

    public function truevoFileLogger($message)
    {
        if (Configuration::get('TRUEVO_DEBUG_LOGGING')) {
            $logger = new FileLogger(0);
            $logger->setFilename($this->getDebugLogFilePath());
            $logger->logDebug($message);
        }
    }

    public function isPaymentMethodEnabled()
    {
        return Configuration::get('TRUEVO_ENABLE_PAYMENT_OPTION') == 1 ? true : false;
    }

    public function getPaymentMethodTitle()
    {
        return $this->l(Configuration::get('TRUEVO_TITLE'));
    }

    public function getPaymentMethodDescription()
    {
        return Configuration::get('TRUEVO_DESC');
    }

    public function isPreauthEnabled()
    {
        return Configuration::get('TRUEVO_ENABLE_PREAUTH') == 1 ? true : false;
    }

    public function isTokenizationEnabled()
    {
        return Configuration::get('TRUEVO_ENABLE_TOKENIZATION') == 1 ? true : false;
    }

    public function getTransactionType()
    {
        return ($this->isPreauthEnabled()) ? 'authorize' : 'sale';
    }

    public function getAuthToken()
    {
        return ($this->isTestMode()) ? Configuration::get('TRUEVO_TEST_AUTH_TOKEN') : Configuration::get('TRUEVO_PROD_AUTH_TOKEN');
    }

    public function getWidgetTheme()
    {
        return Configuration::get('TRUEVO_WIDGET_THEME');
    }

    public function getTruevoLocale()
    {
        return Configuration::get('TRUEVO_WIDGET_LANGUAGE');
    }

    public function showTruevoLogo()
    {
        return Configuration::get('TRUEVO_SHOW_LOGO') == 1 ? true : false;
    }

    public function isTestMode()
    {
        return Configuration::get('TRUEVO_TEST_MODE') == 1 ? true : false;
    }

    public function getMerchantMid()
    {
        return ($this->isTestMode()) ? Configuration::get('TRUEVO_TEST_MID') : Configuration::get('TRUEVO_PROD_MID');
    }

    public function getMerchantTid()
    {
        return ($this->isTestMode()) ? Configuration::get('TRUEVO_TEST_TID') : Configuration::get('TRUEVO_PROD_TID');
    }

    public function getCustomCss()
    {
        return Configuration::get('TRUEVO_CSS_STYLES');
    }

    public function getTruevoWidgetUrl()
    {
        if ($this->isTestMode()) {
            return 'https://sandbox-truevo-drop-in-hosted-payment-widget.s3.eu-west-1.amazonaws.com/merchant.minified.js';
        }

        return Media::getMediaPath(_PS_MODULE_DIR_ . $this->name . '/js/merchant.minified.js');
    }

    public function getTruevoApiUrl()
    {
        return ($this->isTestMode()) ? 'https://sandbox.unify.truevo.com' : 'https://unify.truevo.com';
    }

    public function getFullTruevoEndpoint($endpoint, $queryParams = null)
    {
        $url = $this->getTruevoApiUrl();

        if (strpos($endpoint, '/') != 0) {
            $url .= '/';
        }

        $url .= $endpoint;

        if (!is_null($queryParams)) {
            $url .= '?' . $queryParams;
        }

        return $url;
    }

    public function getDebugLogFilePath()
    {
        return _PS_ROOT_DIR_."/var/logs/truevo-debug.log";
    }

    public function _serializePrecisionFix()
    {
        // json_encode breaks decimals in PHP 7.1 due to the serialize_precision
        if (version_compare(phpversion(), '7.1', '>=')) {
            ini_set( 'serialize_precision', -1 );
        }
    }

    /**
     * Assign response message to the cookie to display on the order page.
     * Gets used in truevopayments.php > displayCaptureOfPurchaseNotification()
     *
     * @param mixed $orderId
     * @param string $type "success" | "info" | "warning" | "danger"
     * @param string $message
     *
     * @return void
     */
    public function assignStatusNotification($orderId, string $type = "success", string $message = "Successfully completed!")
    {
        $this->context->cookie->__set("truevo_api_notification_order_$orderId", json_encode([
            'type'          => $type,
            'message'       => $message,
            'expiry_time' => time() + 5
        ]));

        $this->context->cookie->write();
    }
}
